//  Copyright (c) 2018 Cisco and/or its affiliates.
//
//  Licensed under the Apache License, Version 2.0 (the "License");
//  you may not use this file except in compliance with the License.
//  You may obtain a copy of the License at:
//
//      http://www.apache.org/licenses/LICENSE-2.0
//
//  Unless required by applicable law or agreed to in writing, software
//  distributed under the License is distributed on an "AS IS" BASIS,
//  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
//  See the License for the specific language governing permissions and
//  limitations under the License.

package vpp_ipsec

import (
	"strconv"
	"strings"

	"go.ligato.io/vpp-agent/v3/pkg/models"
)

// ModuleName is the module name used for models.
const ModuleName = "vpp.ipsec"

var (
	ModelSecurityPolicyDatabase models.KnownModel
	ModelSecurityPolicy         models.KnownModel
	ModelSecurityAssociation    models.KnownModel
	ModelTunnelProtection       models.KnownModel
)

func init() {
	// models.Register requires protoreflect capabilities, so we initialize them first
	file_ligato_vpp_ipsec_ipsec_proto_init()

	ModelSecurityPolicyDatabase = models.Register(&SecurityPolicyDatabase{}, models.Spec{
		Module:  ModuleName,
		Version: "v2",
		Type:    "spd",
	}, models.WithNameTemplate("{{.Index}}"))

	ModelSecurityPolicy = models.Register(&SecurityPolicy{}, models.Spec{
		Module:  ModuleName,
		Version: "v2",
		Type:    "sp",
	}, models.WithNameTemplate(
		"spd/{{.SpdIndex}}/"+
			"sa/{{.SaIndex}}/"+
			"{{if .IsOutbound}}outbound/{{else}}inbound/{{end}}"+
			"local-addresses/{{.LocalAddrStart}}-{{.LocalAddrStop}}/"+
			"remote-addresses/{{.RemoteAddrStart}}-{{.RemoteAddrStop}}"))

	ModelSecurityAssociation = models.Register(&SecurityAssociation{}, models.Spec{
		Module:  ModuleName,
		Version: "v2",
		Type:    "sa",
	}, models.WithNameTemplate("{{.Index}}"))

	ModelTunnelProtection = models.Register(&TunnelProtection{}, models.Spec{
		Module:  ModuleName,
		Version: "v2",
		Type:    "tun-protect",
	}, models.WithNameTemplate(
		`{{.Interface}}`+
			`{{if .NextHopAddr}}/nh/{{.NextHopAddr}}{{end}}`,
	))
}

// SPDKey returns the key used in NB DB to store the configuration of the
// given security policy database configuration.
func SPDKey(index uint32) string {
	return models.Key(&SecurityPolicyDatabase{
		Index: index,
	})
}

// SAKey returns the key used in NB DB to store the configuration of the
// given security association configuration.
func SAKey(index uint32) string {
	return models.Key(&SecurityAssociation{
		Index: index,
	})
}

/* SPD <-> interface binding (derived) */
const (
	// spdInterfaceKeyTemplate is a template for (derived) key representing binding
	// between interface and a security policy database.
	spdInterfaceKeyTemplate = "vpp/spd/{spd}/interface/{iface}"
)

const (
	// InvalidKeyPart is used in key for parts which are invalid
	InvalidKeyPart = "<invalid>"
)

/* SPD <-> interface binding (derived) */

// SPDInterfaceKey returns the key used to represent binding between the given interface
// and the security policy database.
func SPDInterfaceKey(spdIndex uint32, ifName string) string {
	if ifName == "" {
		ifName = InvalidKeyPart
	}
	key := strings.Replace(spdInterfaceKeyTemplate, "{spd}", strconv.FormatUint(uint64(spdIndex), 10), 1)
	key = strings.Replace(key, "{iface}", ifName, 1)
	return key
}

// ParseSPDInterfaceKey parses key representing binding between interface and a security
// policy database
func ParseSPDInterfaceKey(key string) (spdIndex string, iface string, isSPDIfaceKey bool) {
	keyComps := strings.Split(key, "/")
	if len(keyComps) >= 5 && keyComps[0] == "vpp" && keyComps[1] == "spd" && keyComps[3] == "interface" {
		iface = strings.Join(keyComps[4:], "/")
		return keyComps[2], iface, true
	}
	return "", "", false
}
